//
//  NTIAPTool.h
//  NTComponent_Example
//
//  Created by tzqiang on 2023/1/11.
//  Copyright © 2023 tzqiang. All rights reserved.
//

#import <Foundation/Foundation.h>

NS_ASSUME_NONNULL_BEGIN

typedef NS_ENUM(NSInteger, NTIAPTransactionError) {
    /// 产品 ID 为空
    NTIAPTransactionErrorProductIdEmpty = 1,
    /// 设备不支持购买
    NTIAPTransactionErrorCannotMakePayments,
    /// 有交易未完成
    NTIAPTransactionErrorTransactionNotFinished,
    /// 获取产品信息失败
    NTIAPTransactionErrorGetProductFailed,
    /// 交易队列为空
    NTIAPTransactionErrorTransactionsEmpty,
    /// 交易失败：SKErrorCode.错误描述，例如（2.用户取消了付款请求）
    NTIAPTransactionErrorTransactionFailed,
    /// 恢复购买失败：SKErrorCode.错误描述，例如（0.未知错误）
    NTIAPTransactionErrorRestoreTransactionFailed
};

@class SKPaymentTransaction;

@interface NTIAPReceiptParams : NSObject

@property (nonatomic, copy) NSString *receipt;
@property (nonatomic, copy) NSString *transactionId;
@property (nonatomic, copy) NSString *productId;

@end

/// 内购工具，version：1.0.2
@interface NTIAPTool : NSObject

/// 是否有交易正在进行
@property (nonatomic, assign, readonly) BOOL isTransactioning;

/// 0-1. 产品购买失败、交易相关失败的回调（当交易失败或者恢复购买失败，errorMessage 返回格式为: SKErrorCode.错误描述），可在此回调中进行埋点统计
@property (nonatomic, copy) void (^failureBlock)(NTIAPTransactionError errorCode, NSString *errorMessage);

/// 0-2 .直接购买、续订、恢复购买等获取票据信息回调，使用交易票据（parmas:  receipt、 transactionId、 productId）与服务器进行校验，可在此回调中进行埋点统计。如果网络调用失败可尝试重试几次，只有校验成功再调用 nt_finishTransaction:
@property (nonatomic, copy) void (^verifyTransactionReceiptHandler)(NTIAPReceiptParams *params, SKPaymentTransaction *transaction);

/// 0-3. 通过 keychain 获取票据信息回调，使用交易票据（parmas:  receipt、 transactionId、 productId）与服务器进行校验，可在此回调中进行埋点统计。失败可尝试重试几次，只有校验成功再调用 nt_deleteKeychainReceipt 删除 keychain 中的票据信息。如果多次尝试还是失败，请把当前票据信息上传服务器或第三方统计平台，当收集日志成功后，再删除 keychain 缓存的交易票据信息。
@property (nonatomic, copy) void (^verifyKeychainReceiptHandler)(NTIAPReceiptParams *params);

+ (instancetype)shareInstance;

/// 当前的交易数组
- (NSArray<SKPaymentTransaction *> *)getTransactions;

/// 0-0. 在 didFinishLaunchingWithOptions 启动添加观察监听，防止掉单漏单
- (void)nt_addIAPObserver;

/// 1. 恢复购买｜恢复订阅，可在回调中进行埋点统计
/// - Parameters:
///   - hudHandler: 恢复购买加载回调
///   - restoreCompletedHandler: 已处理完所有可恢复交易后的回调
- (void)nt_restoreWithHUDHandler:(nullable dispatch_block_t)hudHandler
 completedTransactionsHUDHandler:(nullable dispatch_block_t)restoreCompletedHUDHandler;

/// 2. 直接发起产品购买，获得到票据信息，可在回调中进行埋点统计
/// - Parameters:
///   - productId: 产品 ID
///   - hudHandler: 开始购买加载回调
- (void)nt_buyWithProductId:(NSString *)productId
                 hudHandler:(nullable dispatch_block_t)hudHandler;

/// 3. 结束交易，通过调用方法（1、2），在 verifyTransactionReceiptHandler 回调中进行票据校验成功后，再调用此方法
- (void)nt_finishTransaction:(SKPaymentTransaction *)transaction;

/// 当前是否为沙盒环境
- (BOOL)nt_getIsSandbox;

/// 是否有持久化的 Keychain 交易票据
- (BOOL)nt_hasKeychainReceipt;

/// 4. 使用持久化的 Keychain 校验，可在回调中进行埋点统计
- (void)nt_verifyKeychainReceiptHUDHandler:(nullable dispatch_block_t)hudHandler;

/// 5. 删除持久化的 Keychain 交易票据，通过调用方法（4），在 verifyKeychainReceiptHandler 回调中恢复校验票据成功后，再调用此方法
- (BOOL)nt_deleteKeychainReceipt;

/// 跳转 App Store 订阅管理
- (BOOL)nt_openSubscriptions;

/// 持久化内购业务需要额外的参数信息，可选
/// - Parameter params: 比如项目业务中有商品 ID，优惠券相关的数据，此数据需与持久化的苹果产品 productId 对应关联
- (void)nt_saveToKeychainWithExtraParams:(NSDictionary *)params;

/// 获取持久化的 Keychain 额外数据，可选
- (NSDictionary *)nt_getKeychainWithExtraParams;

/// 删除持久化的 Keychain 额外数据，可选
- (BOOL)nt_deleteKeychainExtraParams;

@end

NS_ASSUME_NONNULL_END
