//
//  NTIAPTool.m
//  NTComponent_Example
//
//  Created by tzqiang on 2023/1/11.
//  Copyright © 2023 tzqiang. All rights reserved.
//

#import "NTIAPTool.h"

#import <StoreKit/StoreKit.h>

#ifndef BLOCK
#define BLOCK(_b_, ...) if(_b_){_b_(__VA_ARGS__);}
#endif

@implementation NTIAPReceiptParams

- (NSString *)description {
    return [NSString stringWithFormat:@"transactionId = %@, productId = %@, receipt = %@", self.transactionId, self.productId, self.receipt];
}

@end

@interface NTIAPTool()
<
SKProductsRequestDelegate,
SKPaymentTransactionObserver
>

@property (nonatomic, copy) NSString *bundleIdentifier;
@property (nonatomic, copy) NSString *IAPReceiptKey;
@property (nonatomic, copy) NSString *IAPTransactionIdKey;
@property (nonatomic, copy) NSString *IAPProductIdKey;
@property (nonatomic, copy) NSString *IAPExtraParamsKey;

/// 当前正在购买交易的产品 ID
@property (nonatomic, copy) NSString *productId;
/// 是否有交易正在进行
@property (nonatomic, assign) BOOL isTransactioning;
@property (nonatomic, copy) dispatch_block_t restoreCompletedHUDHandler;

@end

@implementation NTIAPTool

#pragma mark - instance

+ (instancetype)shareInstance {
    static NTIAPTool *instance;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        instance = [[self alloc] init];
    });
    return instance;
}

- (instancetype)init {
    self = [super init];
    if (self) {
        self.bundleIdentifier = [[NSBundle mainBundle] bundleIdentifier];
        self.IAPReceiptKey = [NSString stringWithFormat:@"%@_IAPReceipt", self.bundleIdentifier];
        self.IAPTransactionIdKey = [NSString stringWithFormat:@"%@_IAPTransactionId", self.bundleIdentifier];
        self.IAPProductIdKey = [NSString stringWithFormat:@"%@_IAPProductId", self.bundleIdentifier];
        self.IAPExtraParamsKey = [NSString stringWithFormat:@"%@_IAPExtraParams", self.bundleIdentifier];
    }
    return self;
}

#pragma mark - public method

- (NSArray<SKPaymentTransaction *> *)getTransactions {
    return [SKPaymentQueue defaultQueue].transactions;
}

- (void)nt_addIAPObserver {
    /*
     只有交易事务被正常结束（主动调用 [[SKPaymentQueue defaultQueue] finishTransaction:transaction]; 完结购买交易）本次支付行为才算完成，
     或者自动订阅续费（续费都 Apple 自动完成的，一般在要过期的前 24 小时开始，苹果会尝试扣费，
     扣费成功的话会在 App 下次启动的时候主动推送给 App。），在每次 App 启动时，通过调用 addTransactionObserver:
     就会触发回调之前所有未完结的事务（paymentQueue:updatedTransactions:）。
     */
    [[SKPaymentQueue defaultQueue] addTransactionObserver:self];
    NSLog(@"-IAP- 已注册支付交易队列观察者");
}

- (void)nt_restoreWithHUDHandler:(nullable dispatch_block_t)hudHandler
 completedTransactionsHUDHandler:(nullable dispatch_block_t)restoreCompletedHUDHandler {
    [self _windowUserInteractionEnabled:NO];
    [[SKPaymentQueue defaultQueue] restoreCompletedTransactions];
    NSLog(@"-IAP- 正在为您恢复购买...");
    BLOCK(hudHandler);
    self.restoreCompletedHUDHandler = restoreCompletedHUDHandler;
}

- (void)nt_buyWithProductId:(NSString *)productId
                 hudHandler:(nullable dispatch_block_t)hudHandler {
    NSLog(@"-IAP- 准备开始购买产品 ID：%@", productId);
    if (productId.length == 0) {
        NSString *errorMessage = @"产品 ID 为空";
        NSLog(@"-IAP- 交易失败：%@", errorMessage);
        BLOCK(self.failureBlock, NTIAPTransactionErrorProductIdEmpty, errorMessage);
        return;
    }
    if (![SKPaymentQueue canMakePayments]) {
        NSString *errorMessage = @"当前设备不支持购买";
        NSLog(@"-IAP- 交易失败：%@", errorMessage);
        BLOCK(self.failureBlock, NTIAPTransactionErrorCannotMakePayments, errorMessage);
        return;
    }
    if ([SKPaymentQueue defaultQueue].transactions.count > 0) {
        NSString *errorMessage = @"当前设备中有购买交易未结束";
        NSLog(@"-IAP- 交易失败：%@", errorMessage);
        BLOCK(self.failureBlock, NTIAPTransactionErrorTransactionNotFinished, errorMessage);
        return;
    }
    NSLog(@"-IAP- 当前设备支持购买");
    BLOCK(hudHandler);
    
    [self _windowUserInteractionEnabled:NO];
    self.productId = productId;
    NSSet *productIs = [NSSet setWithArray:@[productId]];
    SKProductsRequest *request = [[SKProductsRequest alloc] initWithProductIdentifiers:productIs];
    request.delegate = self;
    [request start];
    NSLog(@"-IAP- 正在获取产品信息...");
}

- (BOOL)nt_hasKeychainReceipt {
    return [self _getKeychainReceipt].length != 0 && [self _getKeychainTransactionId].length != 0 && [self _getKeychainProductId].length != 0;
}

- (void)nt_verifyKeychainReceiptHUDHandler:(nullable dispatch_block_t)hudHandler {
    NSLog(@"-IAP- 准备从 keychain 中的获取缓存票据信息");
    if ([self nt_hasKeychainReceipt]) {
        [self _windowUserInteractionEnabled:NO];
        
        BLOCK(hudHandler);
        NTIAPReceiptParams *params = [[NTIAPReceiptParams alloc] init];
        params.receipt = [self _getKeychainReceipt];
        params.transactionId = [self _getKeychainTransactionId];
        params.productId = [self _getKeychainProductId];
        
        NSLog(@"-IAP- keychain 获取票据信息参数 params：%@", [params description]);
        if (params.receipt.length != 0) {
            BLOCK(self.verifyKeychainReceiptHandler, params);
        }
    } else {
        NSLog(@"-IAP- keychain 中没有缓存的票据信息");
    }
}

- (BOOL)nt_deleteKeychainReceipt {
    BOOL isSucceeded1 = [self _deleteKeychainItemWithIdentifier:self.IAPReceiptKey serviceName:self.bundleIdentifier];
    BOOL isSucceeded2 = [self _deleteKeychainItemWithIdentifier:self.IAPTransactionIdKey serviceName:self.bundleIdentifier];
    BOOL isSucceeded3 = [self _deleteKeychainItemWithIdentifier:self.IAPProductIdKey serviceName:self.bundleIdentifier];
    if (isSucceeded1 && isSucceeded2 && isSucceeded3) {
        NSLog(@"-IAP- 已删除 keychain 中的缓存票据信息");
    } else {
        NSLog(@"-IAP- 删除 keychain 中的缓存票据信息失败");
    }
    return isSucceeded1 && isSucceeded2 && isSucceeded3;
}

- (BOOL)nt_openSubscriptions {
    NSURL *url = [NSURL URLWithString:@"https://apps.apple.com/account/subscriptions"];
    if ([[UIApplication sharedApplication] canOpenURL:url]) {
        if (@available(iOS 10.0, *)) {
            [[UIApplication sharedApplication] openURL:url
                                               options:@{ UIApplicationOpenURLOptionsSourceApplicationKey : @YES }
                                     completionHandler:nil];
        } else {
            [[UIApplication sharedApplication] openURL:url];
        }
        return YES;
    }
    return NO;
}

- (void)nt_saveToKeychainWithExtraParams:(NSDictionary *)params {
    if (params.allValues.count > 0) {
        [self _insertKeychainItemWithDictionary:params identifier:self.IAPExtraParamsKey serviceName:self.bundleIdentifier];
        NSLog(@"-IAP- 持久化交易额外参数 params：%@", params);
    }
}

- (NSDictionary *)nt_getKeychainWithExtraParams {
    NSData *params = [self _selectKeychainItemWithIdentifier:self.IAPExtraParamsKey serviceName:self.bundleIdentifier];
    if (params.length == 0) {
        NSLog(@"-IAP- 未获取到持久化交易的额外参数 params");
        return nil;
    }
    
    NSDictionary *dict = [NSJSONSerialization JSONObjectWithData:params options:NSJSONReadingMutableLeaves error:nil];
    NSLog(@"-IAP- 获取持久化交易的额外参数 params：%@", dict);
    return dict;
}

- (BOOL)nt_deleteKeychainExtraParams {
    BOOL isSucceeded = [self _deleteKeychainItemWithIdentifier:self.IAPExtraParamsKey serviceName:self.bundleIdentifier];
    if (isSucceeded) {
        NSLog(@"-IAP- 已删除 keychain 中的内购额外数据");
    } else {
        NSLog(@"-IAP- 删除 keychain 中的内购额外数据失败");
    }
    return isSucceeded;
}

- (BOOL)nt_getIsSandbox {
    BOOL isSandbox = NO;
#if DEBUG
    isSandbox = YES;
#else
    NSString *lastPathComponent = [NSBundle mainBundle].appStoreReceiptURL.lastPathComponent;
    if ([lastPathComponent containsString:@"sandbox"]) {
        // TestFlight
        isSandbox = YES;
    } else {
        isSandbox = NO;
    }
#endif
    return isSandbox;
}

- (void)nt_finishTransaction:(SKPaymentTransaction *)transaction {
    NSLog(@"-IAP- 与服务器校验票据成功，本次购买交易完成");
    [self nt_deleteKeychainReceipt];
    [self _finishTransaction:transaction];
}

#pragma mark - private method

- (void)_windowUserInteractionEnabled:(BOOL)enabled {
    /// 购买过程中是否禁止用户操作 App
    dispatch_async(dispatch_get_main_queue(), ^{
        [UIApplication sharedApplication].keyWindow.userInteractionEnabled = enabled;
        NSLog(@"-IAP- 当前是否允许用户操作 App：%d", (int)enabled);
    });
}

- (NSString *)_getKeychainReceipt {
    NSData *data = [self _selectKeychainItemWithIdentifier:self.IAPReceiptKey serviceName:self.bundleIdentifier];
    if (data.length == 0) {
        return nil;
    }
    
    NSString *receipt = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
    NSLog(@"-IAP- Keychain receipt: %@", receipt);
    return receipt;
}

- (NSString *)_getKeychainTransactionId {
    NSData *data = [self _selectKeychainItemWithIdentifier:self.IAPTransactionIdKey serviceName:self.bundleIdentifier];
    if (data.length == 0) {
        return nil;
    }
    
    NSString *transactionId = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
    NSLog(@"-IAP- Keychain transactionId: %@", transactionId);
    return transactionId;
}

- (NSString *)_getKeychainProductId {
    NSData *data = [self _selectKeychainItemWithIdentifier:self.IAPProductIdKey serviceName:self.bundleIdentifier];
    if (data.length == 0) {
        return nil;
    }
    
    NSString *productId = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
    NSLog(@"-IAP- Keychain productId: %@", productId);
    return productId;
}

- (NSString *)_getErrorMessage:(NSError *)error {
    SKErrorCode errorCode = error.code;
    NSString *detail = @"未知错误";
    switch (errorCode) {
        case SKErrorUnknown:
            detail = @"未知错误";
            break;
        case SKErrorClientInvalid:
            detail = @"不允许 App 执行尝试的操作";
            break;
        case SKErrorPaymentCancelled:
            detail = @"用户取消了付款请求";
            break;
        case SKErrorPaymentInvalid:
            detail = @"App Store 无法识别其中一个支付参数";
            break;
        case SKErrorPaymentNotAllowed:
            detail = @"不允许用户授权付款";
            break;
        case SKErrorStoreProductNotAvailable:
            detail = @"所请求的产品在 App Store 商店中不可用";
            break;
        case SKErrorCloudServicePermissionDenied:
            detail = @"用户不允许访问云服务信息";
            break;
        case SKErrorCloudServiceNetworkConnectionFailed:
            detail = @"设备无法连接到网络";
            break;
        case SKErrorCloudServiceRevoked:
            detail = @"用户已撤销使用此云服务的权限";
            break;
        case SKErrorPrivacyAcknowledgementRequired:
            detail = @"用户尚未接受 Apple 的 Apple Music 隐私政策";
            break;
        case SKErrorUnauthorizedRequestData:
            detail = @"该应用程序正在尝试使用它没有所需权利的属性";
            break;
        case SKErrorInvalidOfferIdentifier:
            detail = @"指定的订阅提供标识符无效";
            break;
        case SKErrorInvalidSignature:
            detail = @"付款折扣中的签名无效";
            break;
        case SKErrorMissingOfferParams:
            detail = @"付款折扣中缺少参数";
            break;
        case SKErrorInvalidOfferPrice:
            detail = @"当前在 App Store Connect 中指定的价格不再有效";
            break;
        case SKErrorOverlayCancelled:
            detail = @"覆盖取消";
            break;
        case SKErrorOverlayInvalidConfiguration:
            detail = @"覆盖配置无效";
            break;
        case SKErrorOverlayTimeout:
            detail = @"覆盖超时";
            break;
        case SKErrorIneligibleForOffer:
            detail = @"用户没有资格享受订阅优惠";
            break;
        case SKErrorUnsupportedPlatform:
            detail = @"当前平台不支持覆盖";
            break;
        case SKErrorOverlayPresentedInBackgroundScene:
            detail = @"客户端尝试在 UIWindowScene 中呈现 SKOverlay，而不是在前台";
            break;
    }
    if ([error.userInfo[NSLocalizedDescriptionKey] length] != 0) {
        detail = error.userInfo[NSLocalizedDescriptionKey];
    }
    
    return [NSString stringWithFormat:@"%ld.%@", errorCode, detail];
}

- (void)_failedTransaction:(SKPaymentTransaction *)transaction {
    self.isTransactioning = NO;
    self.productId = @"";
    
    NSString *errorMessage = [self _getErrorMessage:transaction.error];
    BLOCK(self.failureBlock, NTIAPTransactionErrorTransactionFailed, errorMessage);
    NSLog(@"-IAP- 交易失败：%@", errorMessage);
    
    [self _finishTransaction:transaction];
}

- (void)_restoreTransaction:(SKPaymentTransaction *)transaction {
    NSLog(@"-IAP- 交易恢复处理");
    [self _completeTransaction:transaction];
}

- (void)_completeTransaction:(nullable SKPaymentTransaction *)transaction {
    NTIAPReceiptParams *params = [[NTIAPReceiptParams alloc] init];
    NSURL *appStoreReceiptURL = [[NSBundle mainBundle] appStoreReceiptURL];
    
    if (transaction != nil) {
        NSData *data = [NSData dataWithContentsOfURL:appStoreReceiptURL];
        NSString *receipt = [data base64EncodedStringWithOptions:0];
        [self _insertKeychainItemWithValue:receipt identifier:self.IAPReceiptKey serviceName:self.bundleIdentifier];
        params.receipt = receipt;
        
        NSString *transactionIdentifier = @"";
        if (transaction.transactionState == SKPaymentTransactionStateRestored || transaction.originalTransaction) {
            // 如果是自动续费的订单, originalTransaction 会有内容，或者恢复购买
            transactionIdentifier = transaction.originalTransaction.transactionIdentifier;
        } else {
            // 普通购买，以及第一次购买自动订阅
            transactionIdentifier = transaction.transactionIdentifier;
        }
        [self _insertKeychainItemWithValue:transactionIdentifier identifier:self.IAPTransactionIdKey serviceName:self.bundleIdentifier];
        params.transactionId = transactionIdentifier;
        
        NSString *productId = transaction.payment.productIdentifier;
        [self _insertKeychainItemWithValue:productId identifier:self.IAPProductIdKey serviceName:self.bundleIdentifier];
        params.productId = productId;
        
        NSLog(@"-IAP- 购买交易获取参数 params：%@", [params description]);
        BLOCK(self.verifyTransactionReceiptHandler, params, transaction);
    }
}

- (void)_finishTransaction:(SKPaymentTransaction *)transaction {
    [self _windowUserInteractionEnabled:YES];

    NSLog(@"-IAP- 已结束当前的购买交易");
    // 只有当交易失败，或者客户端支付扣款功并与服务器验证票据成功，才能调用结束完成本次交易
    [[SKPaymentQueue defaultQueue] finishTransaction:transaction];
}

#pragma mark - SKProductsRequestDelegate

- (void)productsRequest:(SKProductsRequest *)request
     didReceiveResponse:(SKProductsResponse *)response {
    NSArray *products = response.products;
    NSLog(@"-IAP- 产品 Product ID:%@", response.invalidProductIdentifiers);
    NSLog(@"-IAP- 产品付费数量: %lu", (unsigned long)products.count);
    
    BOOL isGetProduct = NO;
    for (SKProduct *product in products) {
        NSLog(@"-IAP- 产品详情");
        NSLog(@"-IAP- 产品 ID: %@" , product.productIdentifier);
        NSLog(@"-IAP- 产品标题: %@" , product.localizedTitle);
        NSLog(@"-IAP- 产品描述: %@", product.description);
        NSLog(@"-IAP- 产品本地化描述: %@" , product.localizedDescription);
        NSLog(@"-IAP- 产品价格: %@" , product.price.stringValue);
        
        if ([product.productIdentifier isEqualToString:self.productId]) {
            SKPayment *payment = [SKPayment paymentWithProduct:product];
            [[SKPaymentQueue defaultQueue] addPayment:payment];
            isGetProduct = YES;
            break;
        }
    }
    if (!isGetProduct) {
        [self _windowUserInteractionEnabled:YES];
        
        NSString *errorMessage = @"未获取到产品信息";
        NSLog(@"-IAP- 交易失败：%@", errorMessage);
        BLOCK(self.failureBlock, NTIAPTransactionErrorGetProductFailed, errorMessage);
    }
}

#pragma mark - SKPaymentTransactionObserver

- (void)paymentQueue:(SKPaymentQueue *)queue
 updatedTransactions:(NSArray<SKPaymentTransaction *> *)transactions {
    NSInteger count = [SKPaymentQueue defaultQueue].transactions.count;
    NSLog(@"-IAP- 交易队列更新信息，队列中有 %ld 个交易", count);
    if (count > 0) {
        self.isTransactioning = YES;
        
        [[SKPaymentQueue defaultQueue].transactions enumerateObjectsUsingBlock:^(SKPaymentTransaction * _Nonnull transaction, NSUInteger idx, BOOL * _Nonnull stop) {
            switch (transaction.transactionState) {
                case SKPaymentTransactionStatePurchasing:
                    NSLog(@"-IAP- 交易队列信息: 交易添加到队列中");
                    break;
                case SKPaymentTransactionStatePurchased:
                    NSLog(@"-IAP- 交易队列信息: 交易完成");
                    [self _completeTransaction:transaction];
                    break;
                case SKPaymentTransactionStateFailed:
                    NSLog(@"-IAP- 交易队列信息: 交易失败");
                    [self _failedTransaction:transaction];
                    break;
                case SKPaymentTransactionStateRestored:
                    NSLog(@"-IAP- 交易队列信息: 交易恢复");
                    [self _restoreTransaction:transaction];
                    break;
                case SKPaymentTransactionStateDeferred:
                    NSLog(@"-IAP- 交易队列信息: 交易延迟");
                    break;
            }
        }];
    } else {
        self.isTransactioning = NO;
        [self _windowUserInteractionEnabled:YES];
        
        NSString *errorMessage = @"当前交易队列为空";
        NSLog(@"-IAP- 交易失败：%@", errorMessage);
        BLOCK(self.failureBlock, NTIAPTransactionErrorTransactionsEmpty, errorMessage);
    }
}

- (void)paymentQueueRestoreCompletedTransactionsFinished:(SKPaymentQueue *)queue {
    self.isTransactioning = NO;
    [self _windowUserInteractionEnabled:YES];
    BLOCK(self.restoreCompletedHUDHandler);
    NSLog(@"-IAP- 支付队列处理完所有可恢复交易后调用此方法 paymentQueueRestoreCompletedTransactionsFinished。您的应用程序可以不需要在此方法中执行任何操作。");
}

- (void)paymentQueue:(SKPaymentQueue *)queue restoreCompletedTransactionsFailedWithError:(NSError *)error {
    self.isTransactioning = NO;
    [self _windowUserInteractionEnabled:YES];
    
    NSString *errorMessage = [self _getErrorMessage:error];
    BLOCK(self.failureBlock, NTIAPTransactionErrorRestoreTransactionFailed, errorMessage);
    NSLog(@"-IAP- 恢复购买失败：%@", errorMessage);
}

#pragma mark - Keychain

- (NSData *)_selectKeychainItemWithIdentifier:(NSString *)identifier serviceName:(NSString *)serviceName {
    NSMutableDictionary *dicForSelect = [self _baseAttributeDictionary:identifier serviceName:serviceName];

    [dicForSelect setObject:(id)kSecMatchLimitOne forKey:(id)kSecMatchLimit];
    [dicForSelect setObject:(id)kCFBooleanTrue forKey:(id)kSecReturnData];

    NSData *result = nil;
    OSStatus status = SecItemCopyMatching((CFDictionaryRef)dicForSelect, (void *)&result);
    if (status == errSecSuccess) {
        // success
    } else {
        // failure
    }
    [self _logAction:@"SecItemCopyMatching" status:(NSInteger)status];
    
    return result;
}

- (BOOL)_insertKeychainItemWithValue:(NSString *)value identifier:(NSString *)identifier serviceName:(NSString *)serviceName {
    NSMutableDictionary *dicForInsert = [self _baseAttributeDictionary:identifier serviceName:serviceName];
    NSData *dataForInsert = [value dataUsingEncoding:NSUTF8StringEncoding];
    [dicForInsert setObject:dataForInsert forKey:(id)kSecValueData];
    
    BOOL isSucceeded;
    OSStatus status = SecItemAdd((CFDictionaryRef)dicForInsert, NULL);
    if (status == errSecSuccess) {
        isSucceeded = YES;
    } else {
        isSucceeded = NO;
    }
    [self _logAction:@"SecItemAdd" status:(NSInteger)status];

    return isSucceeded;
}

- (BOOL)_deleteKeychainItemWithIdentifier:(NSString *)identifier serviceName:(NSString *)serviceName {
    NSMutableDictionary *dicForDelete = [self _baseAttributeDictionary:identifier serviceName:serviceName];
    
    BOOL isSucceeded;
    OSStatus status = SecItemDelete((CFDictionaryRef)dicForDelete);
    if (status == errSecSuccess) {
        isSucceeded = YES;
    } else {
        isSucceeded = NO;
    }
    [self _logAction:@"SecItemDelete" status:(NSInteger)status];

    return isSucceeded;
}

- (BOOL)_insertKeychainItemWithDictionary:(NSDictionary *)dict identifier:(NSString *)identifier serviceName:(NSString *)serviceName {
    NSMutableDictionary *dicForInsert = [self _baseAttributeDictionary:identifier serviceName:serviceName];
    NSError *error;
    NSData *dataFromDict = [NSJSONSerialization dataWithJSONObject:dict
                                                           options:0
                                                             error:&error];
    [dicForInsert setObject:dataFromDict forKey:(id)kSecValueData];
    
    BOOL isSucceeded;
    OSStatus status = SecItemAdd((CFDictionaryRef)dicForInsert, NULL);
    if (status == errSecSuccess) {
        isSucceeded = YES;
    } else {
        isSucceeded = NO;
    }
    [self _logAction:@"SecItemAdd" status:(NSInteger)status];

    return isSucceeded;
}

- (NSMutableDictionary *)_baseAttributeDictionary:(NSString *)identifier serviceName:(NSString *)serviceName {
    NSMutableDictionary *baseAttributeDictionary = [[NSMutableDictionary alloc] init];
    [baseAttributeDictionary setObject:(id)kSecClassGenericPassword forKey:(id)kSecClass];
    NSData *encodedIdentifier = [identifier dataUsingEncoding:NSUTF8StringEncoding];
    [baseAttributeDictionary setObject:encodedIdentifier forKey:(id)kSecAttrGeneric];

    [baseAttributeDictionary setObject:encodedIdentifier forKey:(id)kSecAttrAccount];
    [baseAttributeDictionary setObject:serviceName forKey:(id)kSecAttrService];
    
    return baseAttributeDictionary;
}

- (void)_logAction:(NSString *)action status:(NSInteger)status {
    NSLog(@"%@ Executed: `[KeychainAction: %@], [OSStatus: %ld]`", NSStringFromClass(self.class), action ,(long)status);
}

@end
